unit TriangleMain;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  DirectDraw, Direct3D, DDUtil; //ddutil neni standard

type
  TTriangleForm = class(TForm)
    procedure FormActivate(Sender: TObject);
    procedure FormDeactivate(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure FormCreate(Sender: TObject);
    procedure FormKeyUp(Sender: TObject; var Key: Word;
      Shift: TShiftState);
  private
    { Private declarations }
    FActive: Boolean;
    fTimeKey: Integer;
    FDD: IDirectDraw7;
    FDD3D: IDirect3D7;
    FDD3DDev: IDirect3DDevice7;
    FDDSPrimary: IDirectDrawSurface7;
    FDDSBack: IDirectDrawSurface7;
    FDDSZBuffer: IDirectDrawSurface7;
    FDDClipper: IDirectDrawClipper;
    ScreenRect, ViewportRect: TRect;
    dwRenderWidth, dwRenderHeight: DWORD;
    vp: TD3DViewport7;
    TriVert: array[0..5] of TD3DVertex;
    ddpfZBuffer: TDDPixelFormat;
    procedure FormSetCursor(var Msg: TMessage); message WM_SETCURSOR;
    procedure ErrorOut(hRet: hResult; FuncName: string);
    function  InitDirectDraw: Boolean;
    procedure FormOnIdle(Sender: TObject; var Done: Boolean);
    function  ShowFrame: Boolean;
    function  UpdateFrame: Boolean;
    procedure ExitDirectDraw;
    procedure PrepareGeometry;
    procedure Render;
    procedure RestoreSurfaces;
    procedure WMMove(var Message: TWMMove); message WM_MOVE;
  public
     { Public declarations }
  end;

var
  TriangleForm: TTriangleForm;

implementation

{$R *.DFM}

function EnumZBufferCallback(var pddpf: TDDPixelFormat; lpContext: Pointer): hResult; stdcall;
begin
  // For this tutorial, we are only interested in z-buffers, so ignore any
  // other formats (e.g. DDPF_STENCILBUFFER) that get enumerated. An app
  // could also check the depth of the z-buffer (16-bit, etc,) and make a
  // choice based on that, as well. For this tutorial, we'll take the first
  // one we get.
  if pddpf.dwFlags = DDPF_ZBUFFER then
  begin
    TriangleForm.ddpfZBuffer := pddpf;

    // Return with D3DENUMRET_CANCEL to end the search.
    Result := D3DENUMRET_CANCEL;
    Exit;
  end;

  // Return with D3DENUMRET_OK to continue the search.
  Result := D3DENUMRET_OK;
end;

procedure TTriangleForm.ErrorOut(hRet: hResult; FuncName: string);
var s: string;
begin
  s := FuncName + ': ' + #13 + DDErrorString(hRet);
  MessageDlg(s, mtWarning, [mbOk], 0);
end;

procedure TTriangleForm.ExitDirectDraw;
begin
  if Assigned(FDD) then
  begin
    if Assigned(FDD3D) then
    begin
      if Assigned(FDD3DDev) then FDD3DDev := nil;
      FDD3D := nil;
    end;
    if Assigned(FDDSZBuffer) then FDDSZBuffer := nil;
    if Assigned(FDDSBack) then FDDSBack := nil;
    if Assigned(FDDSPrimary) then FDDSPrimary := nil;
    FDD := nil;
  end;
end;

function TTriangleForm.ShowFrame: Boolean;
var hRet: hResult;
begin
  Result := False;

  if FDDSPrimary = nil then
  begin
    ErrorOut(E_FAIL, 'FlipPages');
    Exit;
  end;

  // We are in windowed mode, so perform a blit from the backbuffer to the
  // correct position on the primary surface
  hRet := FDDSPrimary.Blt(@ScreenRect, FDDSBack, @ViewportRect, DDBLT_WAIT, nil);
  if hRet = DDERR_SURFACELOST then RestoreSurfaces;
  if hRet <> DD_OK then
  begin
    ErrorOut(hRet, 'Blt');
    Exit;
  end;

  // OK
  Result := True;
end;

procedure TTriangleForm.FormSetCursor(var Msg: TMessage);
begin
  SetCursor(0);
  Msg.Result := 1;
end;

function TTriangleForm.InitDirectDraw: Boolean;
var ddsd: TDDSurfaceDesc2;
    hRet: hResult;
begin
  Result := False;

  // hlavni directdraw7
  hRet := DirectDrawCreateEx(nil, FDD, IDirectDraw7, nil);
  if hRet <> DD_OK then
  begin
    ErrorOut(hRet, 'DirectDrawCreateEx');
    Exit;
  end;

  // nastav kooperativu
  hRet := FDD.SetCooperativeLevel(Handle, DDSCL_NORMAL);
  if hRet <> DD_OK then
  begin
    ErrorOut(hRet, 'SetCooperativeLevel');
    Exit;
  end;

  // vytvori prim retezec surfacu s 1 backbufferem
  FillChar(ddsd, SizeOf(ddsd), 0);
  ddsd.dwSize := SizeOf(ddsd);
  ddsd.dwFlags := DDSD_CAPS;
  ddsd.ddsCaps.dwCaps := DDSCAPS_PRIMARYSURFACE or DDSCAPS_3DDEVICE;
  hRet := FDD.CreateSurface(ddsd, FDDSPrimary, nil);
  if hRet <> DD_OK then
  begin
    ErrorOut(hRet, 'CreateSurface');
    Exit;
  end;

  ddsd.dwFlags := DDSD_WIDTH or DDSD_HEIGHT or DDSD_CAPS;
  ddsd.ddsCaps.dwCaps := DDSCAPS_OFFSCREENPLAIN or DDSCAPS_3DDEVICE;

  // Set the dimensions of the back buffer. Note that if our window changes
  // size, we need to destroy this surface and create a new one.
  ScreenRect := GetClientRect;
  ViewportRect := GetClientRect;
  ScreenRect.TopLeft := ClientToScreen(ScreenRect.TopLeft);
  ScreenRect.BottomRight := ClientToScreen(ScreenRect.BottomRight);
  ddsd.dwWidth := ScreenRect.Right - ScreenRect.Left;
  ddsd.dwHeight := ScreenRect.Bottom - ScreenRect.Top;

  // Create the backbuffer. The most likely reason for failure is running
  // out of video memory. (A more sophisticated app should handle this.)
  hRet := FDD.CreateSurface(ddsd, FDDSBack, nil);
  if hRet <> DD_OK then
  begin
    ErrorOut(hRet, 'CreateBackBuffer');
    Exit;
  end;

  hRet := FDD.CreateClipper(0, FDDClipper, nil);
  if hRet <> DD_OK then
  begin
    ErrorOut(hRet, 'CreateClipper');
    Exit;
  end;

  // Assign it to the window handle, then set
  // the clipper to the desired surface.
  FDDClipper.SetHWnd(0, Handle);
  FDDSPrimary.SetClipper(FDDClipper);
  if Assigned(FDDClipper) then FDDClipper := nil;

  // Query DirectDraw for access to Direct3D
  hRet := FDD.QueryInterface(IDirect3D7, FDD3D);
  if hRet <> DD_OK then
  begin
    ErrorOut(hRet, 'QueryInterface');
    Exit;
  end;

  // Check the display mode, and
  ddsd.dwSize := SizeOf(TDDSurfaceDesc2);
  hRet := FDD.GetDisplayMode(ddsd);
  if hRet <> DD_OK then
  begin
    ErrorOut(hRet, 'GetDisplayMode');
    Exit;
  end;
  if ddsd.ddpfPixelFormat.dwRGBBitCount <= 8 then
  begin
    ErrorOut(DDERR_INVALIDMODE, 'GetDisplayMode');
    Exit;
  end;

  // Enumerating the ZBuffers
  ddpfZBuffer.dwSize := 0;
  ddsd.ddsCaps.dwCaps := DDSCAPS_VIDEOMEMORY;
  FDD3D.EnumZBufferFormats(IID_IDirect3DHALDevice, @EnumZBufferCallback, nil);
  if SizeOf(TDDPixelFormat) <> ddpfZBuffer.dwSize then
  begin
    ddsd.ddsCaps.dwCaps := DDSCAPS_SYSTEMMEMORY;
    FDD3D.EnumZBufferFormats(IID_IDirect3DRGBDevice, @EnumZBufferCallback, nil);
    if SizeOf(TDDPixelFormat) <> ddpfZBuffer.dwSize then
    begin
      ErrorOut(E_FAIL, 'EnumZBufferFormats');
      Exit;
    end;
  end;

  // Get z-buffer dimensions from the render target
  // Setup the surface desc for the z-buffer.
  ddsd.dwFlags := DDSD_CAPS or DDSD_WIDTH or DDSD_HEIGHT or DDSD_PIXELFORMAT;
  ddsd.ddsCaps.dwCaps := ddsd.ddsCaps.dwCaps or DDSCAPS_ZBUFFER;
  ddsd.dwWidth := ScreenRect.Right - ScreenRect.Left;
  ddsd.dwHeight := ScreenRect.Bottom - ScreenRect.Top;
  ddsd.ddpfPixelFormat := ddpfZBuffer;

  // Create and attach a z-buffer. Real apps should be able to handle an
  // error here (DDERR_OUTOFVIDEOMEMORY may be encountered). For this
  // tutorial, though, we are simply going to exit ungracefully.
  hRet := FDD.CreateSurface(ddsd, FDDSZBuffer, nil);
  if hRet <> DD_OK then
  begin
    ErrorOut(hRet, 'CreateZBuffer');
    Exit;
  end;

  // Attach the z-buffer to the back buffer.
  hRet := FDDSBack.AddAttachedSurface(FDDSZBuffer);
  if hRet <> DD_OK then
  begin
    ErrorOut(hRet, 'AddAttachedSurface');
    Exit;
  end;

  // The GUID here is hard coded. In a real-world application
  // this should be retrieved by enumerating devices.
  hRet := FDD3D.CreateDevice(IID_IDirect3DHALDevice, FDDSBack, FDD3DDev);
  if hRet <> DD_OK then
  begin
    // If the hardware GUID doesn't work, try a software device.
    hRet := FDD3D.CreateDevice(IID_IDirect3DRGBDevice, FDDSBack, FDD3DDev);
    if hRet <> DD_OK then
    begin
      ErrorOut(hRet, 'Create3DDevice');
      Exit;
    end;
  end;

  // Enable z-buffering. (Note: we don't really need to do this every frame.)
  FDD3DDev.SetRenderState(D3DRENDERSTATE_ZENABLE, Cardinal(D3DZB_TRUE));

  dwRenderWidth := ScreenRect.Right - ScreenRect.Left;
  dwRenderHeight := ScreenRect.Bottom - ScreenRect.Top;
  vp.dwX := 0; vp.dwY := 0;
  vp.dwWidth := dwRenderWidth;
  vp.dwHeight := dwRenderHeight;
  vp.dvMinZ := 0; vp.dvMaxZ := 1;

  hRet := FDD3DDev.SetViewport(vp);
  if hRet <> DD_OK then
  begin
    ErrorOut(hRet, 'SetViewPort');
    Exit;
  end;

  PrepareGeometry;

  // OK:
  Result := True;
end;

function TTriangleForm.UpdateFrame: Boolean;
var matSpin: TD3DMatrix;
begin
  // For this simple tutorial, we are rotating the triangle about the y-axis.
  // To do this, just set up a 4x4 matrix defining the rotation, and set it
  // as the new world transform.
  matSpin._11 := 1; matSpin._22 := 1; matSpin._33 := 1; matSpin._44 := 1;
  matSpin._12 := 0; matSpin._13 := 0; matSpin._14 := 0; matSpin._41 := 0;
  matSpin._21 := 0; matSpin._23 := 0; matSpin._24 := 0; matSpin._42 := 0;
  matSpin._31 := 0; matSpin._32 := 0; matSpin._34 := 0; matSpin._43 := 0;

  matSpin._11 := Cos((fTimeKey * 0.5) * Pi / 180);
  matSpin._33 := Cos((fTimeKey * 0.5) * Pi / 180);
  matSpin._13 := -Sin((fTimeKey * 0.5) * Pi / 180);
  matSpin._31 := Sin((fTimeKey * 0.5) * Pi / 180);

  Result := FDD3DDev.SetTransform(D3DTRANSFORMSTATE_WORLD, matSpin) = DD_OK;

  Inc(fTimeKey);
end;

procedure TTriangleForm.FormActivate(Sender: TObject);
begin
  FActive := True;
end;

procedure TTriangleForm.FormDeactivate(Sender: TObject);
begin
  FActive := False;
end;

procedure TTriangleForm.FormShow(Sender: TObject);
begin
  if not InitDirectDraw then Application.Terminate;
end;

procedure TTriangleForm.FormOnIdle(Sender: TObject; var Done: Boolean);
begin
  if FActive then
  begin
    Caption := 'Triangle (Z/Buffer)> frame = ' + IntToStr(fTimeKey);
    UpdateFrame;
    Render;
    ShowFrame;
  end;
  Done := False;
end;

procedure TTriangleForm.FormClose(Sender: TObject; var Action: TCloseAction);
begin
  ExitDirectDraw;
  Application.Minimize;
end;

procedure TTriangleForm.FormCreate(Sender: TObject);
begin
  SetWindowPos(Handle, HWND_TOPMOST, Left, Top, Width, Height, 0);
  fTimeKey := 0;
  Application.OnIdle := FormOnIdle;
end;

procedure TTriangleForm.FormKeyUp(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
  if (Key = VK_ESCAPE) or (Key = VK_F12) then Close;
end;

procedure TTriangleForm.PrepareGeometry;
var p1: TD3DVector;
    p2: TD3DVector;
    p3: TD3DVector;
    vNormal: TD3DVector ;
    mtrl: TD3DMaterial7;
    mat, matWorld, matView, matProj: TD3DMatrix;
begin
  p1.x := 0; p1.y := 3; p1.z := 0;
  p2.x := 3; p2.y := -3; p2.z := 0;
  p3.x := -3; p3.y := -3; p3.z := 0;
  vNormal.x := 0; vNormal.y := 0; vNormal.z := 1;

  // front side
  TriVert[0].x := p1.x; TriVert[0].y := p1.y; TriVert[0].z := p1.z;
  TriVert[0].nx := vNormal.x; TriVert[0].ny := vNormal.y; TriVert[0].nz := vNormal.z;
  TriVert[0].tu := 0; TriVert[0].tv := 0;

  TriVert[1].x := p2.x; TriVert[1].y := p2.y; TriVert[1].z := p2.z;
  TriVert[1].nx := vNormal.x; TriVert[1].ny := vNormal.y; TriVert[1].nz := vNormal.z;
  TriVert[1].tu := 0; TriVert[1].tv := 0;

  TriVert[2].x := p3.x; TriVert[2].y := p3.y; TriVert[2].z := p3.z;
  TriVert[2].nx := vNormal.x; TriVert[2].ny := vNormal.y; TriVert[2].nz := vNormal.z;
  TriVert[2].tu := 0; TriVert[2].tv := 0;

  // back side
  TriVert[3].x := p1.x; TriVert[3].y := p1.y; TriVert[3].z := p1.z;
  TriVert[3].nx := -vNormal.x; TriVert[3].ny := -vNormal.y; TriVert[3].nz := -vNormal.z;
  TriVert[3].tu := 0; TriVert[3].tv := 0;

  TriVert[4].x := p3.x; TriVert[4].y := p3.y; TriVert[4].z := p3.z;
  TriVert[4].nx := -vNormal.x; TriVert[4].ny := -vNormal.y; TriVert[4].nz := -vNormal.z;
  TriVert[4].tu := 0; TriVert[4].tv := 0;

  TriVert[5].x := p2.x; TriVert[5].y := p2.y; TriVert[5].z := p2.z;
  TriVert[5].nx := -vNormal.x; TriVert[5].ny := -vNormal.y; TriVert[5].nz := -vNormal.z;
  TriVert[5].tu := 0; TriVert[5].tv := 0;

  // set material
  FillChar(mtrl, SizeOf(mtrl), 0);
  mtrl.ambient.r := 1;
  mtrl.ambient.g := 1;
  mtrl.ambient.b := 0;
  FDD3DDev.SetMaterial(mtrl);

  // The ambient lighting value is another state to set. Here, we are turning
  // ambient lighting on to full white.
  FDD3DDev.SetRenderState(D3DRENDERSTATE_AMBIENT, $ffffffff);

  // Start by setting up an identity matrix.
  mat._11 := 1; mat._22 := 1; mat._33 := 1; mat._44 := 1;
  mat._12 := 0; mat._13 := 0; mat._14 := 0; mat._41 := 0;
  mat._21 := 0; mat._23 := 0; mat._24 := 0; mat._42 := 0;
  mat._31 := 0; mat._32 := 0; mat._34 := 0; mat._43 := 0;

  // The world matrix controls the position and orientation
  // of the polygons in world space. We'll use it later to
  // spin the triangle.
  matWorld := mat;
  FDD3DDev.SetTransform(D3DTRANSFORMSTATE_WORLD, matWorld);

  // The view matrix defines the position and orientation of
  // the camera. Here, we are just moving it back along the z-
  // axis by 10 units.
  matView := mat;
  matView._43 := 10;
  FDD3DDev.SetTransform(D3DTRANSFORMSTATE_VIEW, matView);

  // The projection matrix defines how the 3-D scene is "projected"
  // onto the 2-D render target surface. For more information,
  // see "What Is the Projection Transformation?"
  // Set up a very simple projection that scales x and y
  // by 2, and translates z by -1.0.
  matProj := mat;
  matProj._11 := 2;
  matProj._22 := 2;
  matProj._34 := 1;
  matProj._43 := -1;
  matProj._44 := 0;
  FDD3DDev.SetTransform(D3DTRANSFORMSTATE_PROJECTION, matProj);
end;

procedure TTriangleForm.Render;
var hRet: hResult;
    ViewRect: TD3DRect;
begin
  ViewRect.x1 := 0; ViewRect.y1 := 0; ViewRect.x2 := Width; ViewRect.y2 := Height;

  //HRESULT App_Render( LPDIRECT3DDEVICE7 pd3dDevice, D3DRECT* prcViewRect )
  // Clear the viewport to a blue color.
  FDD3DDev.Clear(1, @ViewRect, D3DCLEAR_TARGET or D3DCLEAR_ZBUFFER, $000000ff, 1, 0);

  // Begin the scene
  hRet := FDD3DDev.BeginScene;
  if hRet <> DD_OK then
  begin
    ErrorOut(hRet, 'BeginScene');
    Exit;
  end;

  // Draw the triangle using a DrawPrimitive() call. Subsequent
  // tutorials will go into more detail on the various calls for
  // drawing polygons.
  hRet := FDD3DDev.DrawPrimitive(D3DPT_TRIANGLELIST, D3DFVF_VERTEX, TriVert, 6, D3DDP_WAIT);
  if hRet <> DD_OK then
  begin
    ErrorOut(hRet, 'DrawPrimitive');
    Exit;
  end;

  // End the scene.
  FDD3DDev.EndScene;
end;

procedure TTriangleForm.RestoreSurfaces;
begin
  if Assigned(FDDSPrimary) and Bool(FDDSPrimary.IsLost) then FDDSPrimary._Restore;
  if Assigned(FDDSBack) and Bool(FDDSBack.IsLost) then FDDSBack._Restore;
  if Assigned(FDDSZBuffer) and Bool(FDDSZBuffer.IsLost) then FDDSZBuffer._Restore;
end;

procedure TTriangleForm.WMMove(var Message: TWMMove);
var dwWidth, dwHeight: DWORD;
begin
  dwWidth := ScreenRect.Right - ScreenRect.Left;
  dwHeight := ScreenRect.Bottom - ScreenRect.Top;
  SetRect(ScreenRect, Message.XPos, Message.YPos, Message.XPos + dwWidth, Message.YPos + dwHeight );
end;

end.
